/* ------------------ **
**      Includes      **
** ------------------ */

#include "Yap.h"
#if defined(YAPOR) || defined(TABLING)
#include <unistd.h>
#include <signal.h>
#ifdef TABLING
#include "tab.macros.h"
#endif /* TABLING */



/* ------------------------------------- **
**      Local functions declaration      **
** ------------------------------------- */

static void receive_signals(int s);



/* -------------------------- **
**      Global functions      **
** -------------------------- */

void init_global(int n_workers, int sch_loop, int delay_load) {
  int i;

#if defined(YAPOR_STATISTICS) || defined(TABLING_STATISTICS)
  /* global data related to statistics */
  INIT_LOCK(GLOBAL_STATS_lock);
#ifdef YAPOR_STATISTICS
  GLOBAL_STATS_allocated_or_fr = 0;
  GLOBAL_STATS_used_or_fr = 0;
  GLOBAL_STATS_allocated_ans_fr = 0;
  GLOBAL_STATS_used_ans_fr = 0;
  GLOBAL_STATS_allocated_sol_fr = 0;
  GLOBAL_STATS_used_sol_fr = 0;
#endif /* YAPOR_STATISTICS */
#ifdef TABLING_STATISTICS
  GLOBAL_STATS_allocated_tab_ent = 0;
  GLOBAL_STATS_used_tab_ent = 0;
  GLOBAL_STATS_allocated_sg_node = 0;
  GLOBAL_STATS_used_sg_node = 0;
  GLOBAL_STATS_allocated_sg_fr = 0;
  GLOBAL_STATS_used_sg_fr = 0;
  GLOBAL_STATS_allocated_ans_node = 0;
  GLOBAL_STATS_used_ans_node = 0;
  GLOBAL_STATS_allocated_dep_fr = 0;
  GLOBAL_STATS_used_dep_fr = 0;
#endif /* TABLING_STATISTICS */
#ifdef OPTYAP_STATISTICS
  GLOBAL_STATS_allocated_susp_fr = 0;
  GLOBAL_STATS_used_susp_fr = 0;
#endif /* OPTYAP_STATISTICS */
#endif /* YAPOR_STATISTICS || TABLING_STATISTICS */

#ifdef YAPOR
  /* global data related to memory management */
#if defined(YAPOR_ERRORS) || defined(TABLING_ERRORS) || defined(YAPOR_DEBUG) || defined(TABLING_DEBUG)
  INIT_LOCK(GLOBAL_LOCKS_stderr_messages);
#endif /* YAPOR_ERRORS || TABLING_ERRORS || YAPOR_DEBUG || TABLING_DEBUG */
  GLOBAL_LOCKS_who_locked_heap = MAX_WORKERS;
  INIT_LOCK(GLOBAL_LOCKS_heap_access);
  INIT_LOCK(GLOBAL_LOCKS_bm_idle_workers);
  INIT_LOCK(GLOBAL_LOCKS_bm_root_cp_workers);
  INIT_LOCK(GLOBAL_LOCKS_bm_invisible_workers);
  INIT_LOCK(GLOBAL_LOCKS_bm_requestable_workers);
  INIT_LOCK(GLOBAL_LOCKS_bm_executing_workers);
  INIT_LOCK(GLOBAL_LOCKS_bm_finished_workers);
  INIT_LOCK(GLOBAL_LOCKS_alloc_area);
  INIT_LOCK(GLOBAL_LOCKS_next_free_or_fr);
  INIT_LOCK(GLOBAL_LOCKS_next_free_ans_fr);
  INIT_LOCK(GLOBAL_LOCKS_next_free_sol_fr);
#ifdef TABLING
  INIT_LOCK(GLOBAL_LOCKS_next_free_tab_ent);
  INIT_LOCK(GLOBAL_LOCKS_next_free_sg_node);
  INIT_LOCK(GLOBAL_LOCKS_next_free_sg_fr);
  INIT_LOCK(GLOBAL_LOCKS_next_free_ans_node);
  INIT_LOCK(GLOBAL_LOCKS_next_free_dep_fr);
  INIT_LOCK(GLOBAL_LOCKS_next_free_susp_fr);
#endif /* TABLING */

  /* global static data */
  number_workers = n_workers;
  worker_pid(0) = getpid();
  for (i = 1; i < number_workers; i++) worker_pid(i) = 0;
  SCHEDULER_LOOP = sch_loop;
  DELAYED_RELEASE_LOAD = delay_load;

  /* global data related to or-performance */
  GLOBAL_number_goals = 0;
  GLOBAL_best_times(0) = 0;
  GLOBAL_performance_mode = PERFORMANCE_OFF;

  /* global data related to or-parallelism */
  BITMAP_clear(GLOBAL_bm_present_workers);
  for (i = 0; i < number_workers; i++) 
    BITMAP_insert(GLOBAL_bm_present_workers, i);
  BITMAP_copy(GLOBAL_bm_idle_workers, GLOBAL_bm_present_workers);
  BITMAP_clear(GLOBAL_bm_root_cp_workers);
  BITMAP_clear(GLOBAL_bm_invisible_workers);
  BITMAP_clear(GLOBAL_bm_requestable_workers);
  BITMAP_clear(GLOBAL_bm_executing_workers);
  BITMAP_copy(GLOBAL_bm_finished_workers, GLOBAL_bm_present_workers);
  if (number_workers == 1)
    GLOBAL_execution_mode = SEQUENTIAL_MODE;
  else
    GLOBAL_execution_mode = PARALLEL_MODE;
#endif /* YAPOR */

#ifdef TABLING
  /* global data related to tabling */
  for (i = 0; i < MAX_TABLE_VARS; i++)
    GLOBAL_table_var_enumerator(i) = (CELL) & GLOBAL_table_var_enumerator(i);
#endif /* TABLING */

  return;
}


void init_local(void) {
#ifdef YAPOR
  /* local data related to or-parallelism */
  LOCAL = REMOTE[worker_id];
  LOCAL_top_cp = B_BASE;
  LOCAL_top_or_fr = GLOBAL_root_or_fr;
  LOCAL_load = - DELAYED_RELEASE_LOAD;
  LOCAL_request = MAX_WORKERS;
  INIT_LOCK(LOCAL_lock_signals);
  LOCAL_cut_cp = NULL;
#endif /* YAPOR */
  INIT_LOCK(LOCAL_lock);
#ifdef TABLING
  /* local data related to tabling */
#ifdef YAPOR
  LOCAL_top_cp_on_stack = B_BASE; /* ??? */
  LOCAL_top_collected_susp_fr = NULL;
#endif /* YAPOR */
  LOCAL_top_dep_fr = GLOBAL_root_dep_fr; 
  LOCAL_top_sg_fr = NULL; 
#endif /* TABLING */
  return;
}


void make_root_frames(void) {
#ifdef YAPOR
  or_fr_ptr or_fr;
#endif  /* YAPOR */
#ifdef TABLING
  dep_fr_ptr dep_fr;
#endif /* TABLING */

#ifdef YAPOR
  /* root or frame */
  ALLOC_OR_FRAME(or_fr);   
  INIT_LOCK(OrFr_lock(or_fr));
  OrFr_alternative(or_fr) = NULL;
  BITMAP_copy(OrFr_members(or_fr), GLOBAL_bm_present_workers);
  OrFr_node(or_fr) = NULL;
  OrFr_nearest_livenode(or_fr) = NULL;
  OrFr_depth(or_fr) = 0;
  OrFr_nearest_leftnode(or_fr) = NULL;
  OrFr_pend_cut_cp(or_fr) = NULL;
  OrFr_solutions(or_fr) = NULL;
#ifdef TABLING
  BITMAP_clear(OrFr_hidden_members(or_fr));
  OrFr_next_on_stack(or_fr) = NULL;
  OrFr_suspensions(or_fr) = NULL;
#endif /* TABLING */
  OrFr_next(or_fr) = NULL;
  GLOBAL_root_or_fr = or_fr;
#endif /* YAPOR */
#ifdef TABLING
  /* root dependency frame */
  ALLOC_DEPENDENCY_FRAME(dep_fr);
  INIT_LOCK(DepFr_lock(dep_fr));
  DepFr_cons_cp(dep_fr) = CONS_CP(B_BASE);
  DepFr_leader_cp(dep_fr) = NULL ;
  DepFr_sg_fr(dep_fr) = NULL;
  DepFr_last_ans(dep_fr) = NULL;
  DepFr_h_fz(dep_fr) = H_BASE;
  DepFr_b_fz(dep_fr) = B_BASE;
  DepFr_tr_fz(dep_fr) = TR_BASE;
  DepFr_backchain_cp(dep_fr) = NULL;
  DepFr_next(dep_fr) = NULL;
  GLOBAL_root_dep_fr = dep_fr;
#endif /* TABLING */
}


#ifdef YAPOR
void init_workers(void) {
  int proc;
#ifdef ACOW
  int son;
  son = fork();
  if (son == -1) abort_optyap("fork error in function init_workers");
  if (son > 0) {
    struct sigaction sigact;
    GLOBAL_master_worker = getpid();
    sigact.sa_handler = SIG_DFL;
    sigemptyset(&sigact.sa_mask);
    sigact.sa_flags = SA_RESTART;
    sigaction(SIGINT, &sigact, NULL);
    pause();  /* wait SIGINT signal */
    exit(0);
  } else worker_pid(0) = getpid();
#endif /* ACOW */
  for (proc = 1; proc < number_workers; proc++) {
    int son;
    son = fork();
    if (son == -1) abort_optyap("fork error in function init_workers");
    if (son == 0) { 
      /* new worker */
      worker_id = proc;
      remap_memory();
      break;
    }
    else worker_pid(proc) = son;
  }
}
#endif /* YAPOR */


void init_signals(void) {

return;

  signal(SIGQUIT, receive_signals);
  signal(SIGTERM, receive_signals);
  signal(SIGSEGV, receive_signals);
  signal(SIGABRT, receive_signals);
  signal(SIGFPE, receive_signals);
  signal(SIGHUP, receive_signals);
  signal(SIGINT, receive_signals);
  signal(SIGTSTP, receive_signals);
  return;
}



/* ------------------------- **
**      Local functions      **
** ------------------------- */

static void receive_signals(int s) {
  abort_optyap("receiving signal number %d", s);
  return;
}
#endif /* YAPOR || TABLING */
